/** M6280: portable 6280 emulator ****************************/
/**                                                         **/
/**                         M6280.h                         **/
/**                                                         **/
/** This file contains declarations relevant to emulation   **/
/** of 6280 CPU.                                            **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996                      **/
/**               Alex Krasivsky  1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/   
/**     changes to this file.                               **/
/*************************************************************/
#ifndef M6280_H
#define M6280_H

/*
 * this file has been modified for use with the cal.
 */

/* $Id: m6280.h,v 1.2 1999/08/07 16:19:05 nyef Exp $ */

#include "cal.h"
#include "mtypes.h"

                               /* Compilation options:       */
/* #define FAST_RDOP */        /* Separate Op6280()/Rd6280() */
/* #define DEBUG */            /* Compile debugging version  */
/* #define LSB_FIRST */        /* Compile for low-endian CPU */

                               /* Loop6280() returns:        */
#define INT_NONE  0xFFFF       /* No interrupt required      */
#define INT_IRQ	  0x0038       /* Standard IRQ interrupt     */
#define INT_NMI	  0x0066       /* Non-maskable interrupt     */
#define INT_QUIT  0xFFFE       /* Exit the emulation         */

                               /* 6280 status flags:         */
#define	M6280_C_FLAG	  0x01         /* 1: Carry occured           */
#define	M6280_Z_FLAG	  0x02         /* 1: Result is zero          */
#define	M6280_I_FLAG	  0x04         /* 1: Interrupts disabled     */
#define	M6280_D_FLAG	  0x08         /* 1: Decimal mode            */
#define	M6280_B_FLAG	  0x10         /* Break [0 on stk after int] */
#define	M6280_R_FLAG	  0x20         /* Always 1                   */
#define	M6280_V_FLAG	  0x40         /* 1: Overflow occured        */
#define	M6280_N_FLAG	  0x80         /* 1: Result is negative      */

/** Simple Datatypes *****************************************/
/** NOTICE: sizeof(byte)=1 and sizeof(word)=2               **/
/*************************************************************/
/* typedef unsigned char byte; */
/* typedef unsigned short word; */
/* typedef signed char offset; */

/** Structured Datatypes *************************************/
/** NOTICE: #define LSB_FIRST for machines where least      **/
/**         signifcant byte goes first.                     **/
/*************************************************************/
/* typedef union */
/* { */
/*   word W; */
/* #ifdef LSB_FIRST */
/*   struct { byte l,h; } B; */
/* #else */
/*   struct { byte h,l; } B; */
/* #endif */
/* } pair; */

typedef struct M6280
{
  byte A,P,X,Y,S;     /* CPU registers and program counter   */
  pair PC;

  int IPeriod,ICount; /* Set IPeriod to number of CPU cycles */
                      /* between calls to Loop6280()         */
  word IRequest;      /* Set to the INT_IRQ when pending IRQ */
  byte AfterCLI;      /* Private, don't touch                */
  int IBackup;        /* Private, don't touch                */
  cal_cpu User;       /* Arbitrary user data (ID,RAM*,etc.)  */
  byte TrapBadOps;    /* Set to 1 to warn of illegal opcodes */
  word Trap;          /* Set Trap to address to trace from   */
  byte Trace;         /* Set Trace=1 to start tracing        */
    byte memshift;
    word memmask;
    memread8_t *readtable;
    memwrite8_t *writetable;
    byte *page_0;
    byte mpr[8];
} M6280;

/** Reset6280() **********************************************/
/** This function can be used to reset the registers before **/
/** starting execution with Run6280(). It sets registers to **/
/** their initial values.                                   **/
/*************************************************************/
void Reset6280(register M6280 *R);

/** Int6280() ************************************************/
/** This function will generate interrupt of a given type.  **/
/** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
/** will cause a normal interrupt, unless M6280_I_FLAG set in R.  **/
/*************************************************************/
void Int6280(register M6280 *R,register word Type);

/** Run6280() ************************************************/
/** This function will run 6280 code until Loop6280() call  **/
/** returns INT_QUIT. It will return the PC at which        **/
/** emulation stopped, and current register values in R.    **/
/*************************************************************/
word Run6280(register M6280 *R);

/** Debug6280() **********************************************/
/** This function should exist if DEBUG is #defined. When   **/
/** Trace!=0, it is called after each command executed by   **/
/** the CPU, and given the 6280 registers. Emulation exits  **/
/** if Debug6280() returns 0.                               **/
/*************************************************************/
byte Debug6280(register M6280 *R);

/** Loop6280() ***********************************************/
/** 6280 emulation calls this function periodically to      **/
/** check if the system hardware requires any interrupts.   **/
/** This function must return one of following values:      **/
/** INT_NONE, INT_IRQ, INT_NMI, or INT_QUIT to exit the     **/
/** emulation loop.                                         **/
/************************************ TO BE WRITTEN BY USER **/
byte Loop6280(register M6280 *R);

#endif /* M6280_H */

/*
 * $Log: m6280.h,v $
 * Revision 1.2  1999/08/07 16:19:05  nyef
 * removed definition of LSB_FIRST
 *
 * Revision 1.1  1999/08/02 00:19:21  nyef
 * Initial revision
 *
 */
