/*
 * sg1000.c
 *
 * Sega Game 1000 emulation.
 */

/* $Id: sg1000.c,v 1.5 1999/12/05 19:47:41 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
#include "tms9918.h"
#include "sms_psg.h"

#define SG1K_RAMSIZE 0x2000

unsigned char *sg1k_memory;
cal_cpu sg1k_cpu;
tms9918 sg1k_vdp;

unsigned char *sg1k_rombase;
rom_file sg1k_romfile;
int sg1k_romsize;

const struct joypad_template sg1k_joypad_template_pause = {
    7, {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x80} /* has pause button */
};

const struct joypad_template sg1k_joypad_template = {
    6, {0x01, 0x02, 0x04, 0x08, 0x10, 0x20}
};

struct joypad sg1k_joypad_1 = {
    &sg1k_joypad_template_pause, 0
};

struct joypad sg1k_joypad_2 = {
    &sg1k_joypad_template, 0
};

ranged_mmu sg1k_mmu[] = {
    {0x0000, 0x0000, NULL, 0, 0},
    {0x0000, 0x9fff, NULL, 0, 0},
    {0xa000, 0xbfff, NULL, 0, 0},
    {0xc000, 0xdfff, NULL, 0, 0},
    {0xe000, 0xffff, NULL, 0, 0},
};

void sg1k_init_mmu(void)
{
    sg1k_mmu[0].data = sg1k_rombase;
    sg1k_mmu[0].end = sg1k_romsize - 1;
    sg1k_mmu[1].start = sg1k_romsize;
    sg1k_mmu[2].data = sg1k_memory - 0xa000;
    sg1k_mmu[3].data = sg1k_memory - 0xc000;
    sg1k_mmu[4].data = sg1k_memory - 0xe000;
}

unsigned char sg1k_read(cal_cpu cpu, unsigned long addr)
{
    if (addr < sg1k_romsize) {
	return sg1k_rombase[addr];
    } else if (addr < 0xa000) {
	return 0x00;
    } else {
	return sg1k_memory[addr & 0x1fff];
    }
}

void sg1k_write(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (addr >= 0xa000) {
	sg1k_memory[addr & 0x1fff] = data;
    }
}

memread8_t sg1k_readfunc = &sg1k_read;
memwrite8_t sg1k_writefunc = &sg1k_write;

unsigned char sg1k_io_read(cal_cpu cpu, unsigned long address)
{
    if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
	    return tms9918_readport1(sg1k_vdp);
	} else {
	    return tms9918_readport0(sg1k_vdp);
	}
    } else if (((address & 0xfe) == 0xdc) || ((address & 0xfe) == 0xc0)) {
	if (address & 1) {
	    return (sg1k_joypad_2.data >> 2) ^ 0xff;
	} else {
	    ui_update_joypad(&sg1k_joypad_1);
	    ui_update_joypad(&sg1k_joypad_2);
	    return (((sg1k_joypad_1.data) & 0x3f) | (sg1k_joypad_2.data << 6)) ^ 0xff;
	}
    } else if ((address & 0xfe) == 0xde) {
	/* FIXME: What goes here? */
    } else {
	deb_printf("sg1k: I/O read 0x%02x.\n", address & 0xff);
    }

    return 0;
}

void sg1k_io_write(cal_cpu cpu, unsigned long address, unsigned char data)
{
    if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
	    tms9918_writeport1(sg1k_vdp, data);
	} else {
	    tms9918_writeport0(sg1k_vdp, data);
	}
    } else if ((address & 0xfe) == 0x7e) {
	sms_psg_write(data);
    } else if ((address & 0xfe) == 0xde) {
	/* FIXME: What goes here? */
    } else {
	deb_printf("sg1k: I/O write 0x%02x = 0x%02x.\n", address & 0xff, data);
    }
}

void sg1k_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: this is .5 over what it should be. */
	cpu->run(cpu);
	if (tms9918_periodic(sg1k_vdp)) {
	    cpu->irq(cpu, 0xff);
	}
	if (i == 42) { /* Somewhat arbitrary, but it will do */
	    /* NOTE: the use of the otherwise unused 0x40 bit is a hack */
	    ui_update_joypad(&sg1k_joypad_1);
	    if (sg1k_joypad_1.data & 0x80) {
		if (!(sg1k_joypad_1.data & 0x40)) {
		    cpu->nmi(cpu);
		}
		sg1k_joypad_1.data |= 0x40;
	    } else {
		sg1k_joypad_1.data &= ~0x40;
	    }
	}
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	    system_flags |= F_QUIT;
	}
    }
}

void sg1k_init_cpu(void)
{
    sg1k_cpu = cal_create(CPUT_MARATZ80);
    sg1k_cpu->setmmu0(sg1k_cpu, sg1k_mmu);
    sg1k_cpu->setmmu8(sg1k_cpu, 0, 0, &sg1k_readfunc, &sg1k_writefunc);
    sg1k_cpu->setiou(sg1k_cpu, sg1k_io_read, sg1k_io_write);
    sg1k_cpu->reset(sg1k_cpu);
}

void sg1k_init_joypads(void)
{
    if (ui_register_joypad(&sg1k_joypad_1)) {
	deb_printf("sg1k_run: joypad 1... registered.\n");
    } else {
	deb_printf("sg1k_run: joypad 1... not registered.\n");
    }
    
    if (ui_register_joypad(&sg1k_joypad_2)) {
	deb_printf("sg1k_run: joypad 2... registered.\n");
    } else {
	deb_printf("sg1k_run: joypad 2... not registered.\n");
    }
}

void sg1k_run(rom_file romfile)
{
    sg1k_memory = calloc(1, SG1K_RAMSIZE);
    
    if (!sg1k_memory) {
	deb_printf("out of memory.\n");
    }

    sg1k_rombase = romfile->data;
    sg1k_romsize = romfile->size;

    sg1k_init_mmu();
    
    sg1k_init_cpu();
    
    sg1k_vdp = tms9918_create();
    
    sms_psg_init();

    sg1k_init_joypads();

    dn_shutdown = sms_psg_done; /* NOTE: cheating */
    
    set_timeslice((void (*)(void *)) &sg1k_timeslice, sg1k_cpu);
}


/*
 * $Log: sg1000.c,v $
 * Revision 1.5  1999/12/05 19:47:41  nyef
 * fixed to shutdown the PSG when exiting
 *
 * Revision 1.4  1999/12/04 01:03:52  nyef
 * cleaned up sg1k_run() and friends
 *
 * Revision 1.3  1999/11/28 16:59:52  nyef
 * cleared out some vestigial code (old CPU interface)
 *
 * Revision 1.2  1999/11/27 20:45:01  nyef
 * disabled debug output on reads from I/O ports 0xde and 0xdf
 *
 * Revision 1.1  1999/11/27 20:07:56  nyef
 * Initial revision
 *
 */
